"""This module contains the classes for building a dependency graph from an ASP document."""

from __future__ import annotations

from dataclasses import dataclass

from mkdocstrings_handlers.asp.document import Document
from mkdocstrings_handlers.asp.semantics.predicate import Predicate


@dataclass
class DependencyGraphNode:
    """Dependency graph node."""

    signature: str
    """The predicate this node represents."""
    positive: set[str]
    """The positive dependencies of the predicate."""
    negative: set[str]
    """The negative dependencies of the predicate."""


@dataclass
class DependencyGraph:
    """Dependency graph of an ASP document."""

    nodes: list[DependencyGraphNode]
    """The nodes of the dependency graph."""

    @staticmethod
    def from_document(documents: list[Document]) -> "DependencyGraph":
        """
        Create a dependency graph from a list of ASP documents.

        Args:
            documents: The list of documents to create the graph from.

        Returns:
            The dependency graph.
        """
        nodes: dict[str, DependencyGraphNode] = {}

        # Find all provided predicates and their dependencies
        for document in documents:
            for statement in document.statements:
                for predicate, _ in statement.provided_predicates:
                    signature = predicate.signature
                    if signature not in nodes:
                        nodes[signature] = DependencyGraphNode(signature, set(), set())

                    # Add positive and negative dependencies from needed_predicates
                    positive = {p.signature for p, n in statement.needed_predicates if not n}
                    negative = {p.signature for p, n in statement.needed_predicates if n}
                    nodes[signature].positive.update(positive)
                    nodes[signature].negative.update(negative)

        # Find all predicates that are not provided but are used in the document
        # these predicates are considered inputs
        for document in documents:
            for signature, predicate in document.predicates.items():
                if signature not in nodes:
                    nodes[signature] = DependencyGraphNode(signature, set(), set())
                    predicate.is_input = True

        return DependencyGraph(list(nodes.values()))



""" This module contains the LineComment class, which represents a single line comment in an ASP document."""

from __future__ import annotations

from dataclasses import dataclass

from tree_sitter import Node


@dataclass
class LineComment:
    """A line comment in an ASP document."""

    row: int
    """ The row of the line comment. """
    line: str
    """ The line of text of the comment. """

    @staticmethod
    def from_node(node: Node) -> LineComment:
        """
        Create a line comment from a node.
        """
        clean_text = node.text.decode("utf-8").removeprefix("%").strip()

        return LineComment(row=node.start_point.row, line=clean_text)



""" This module contains the function to traverse a tree-sitter tree in a pre-order depth-first manner. """

from typing import Callable

from tree_sitter import Node, Tree


def traverse(tree: Tree, on_enter: Callable[[Node], None], on_exit: Callable[[Node], None]) -> None:
    """
    Traverse the given tree in pre-order depth-first manner and call the given functions on each node.

    Args:
        tree: The tree to traverse.
        on_enter: The function to call when entering a node.
        on_exit: The function to call when exiting a node.

    Returns:
        None
    """
    cursor = tree.walk()

    while True:
        # Visit node the first time
        on_enter(cursor.node)

        # Travel down as far as possible
        if cursor.goto_first_child():
            continue

        # If the lowest node is reached, traverse to sibling
        if cursor.goto_next_sibling():
            continue

        # If neither is possible, traverse back up
        while True:
            # Stop traversing if root node is reached
            if not cursor.goto_parent():
                return

            on_exit(cursor.node)

            # Since we've already visited the parent on the way down,
            # try to traverse to sibling and start to move down again
            if cursor.goto_next_sibling():
                break



"""
Module containing the handler for ASP files.
"""

from collections import deque
from pathlib import Path
from typing import Any

from markupsafe import Markup
from mkdocstrings.handlers.base import BaseHandler
from mkdocstrings.handlers.rendering import HeadingShiftingTreeprocessor

from mkdocstrings_handlers.asp.document import Document
from mkdocstrings_handlers.asp.features.dependency_graph import DependencyGraph
from mkdocstrings_handlers.asp.features.encoding_info import EncodingInfo
from mkdocstrings_handlers.asp.features.predicate_info import PredicateInfo
from mkdocstrings_handlers.asp.semantics.document_parser import DocumentParser
from mkdocstrings_handlers.asp.tree_sitter.parser import ASPParser

try:
    import tomllib  # Python 3.11+
except ModuleNotFoundError:
    import tomli as tomllib  # Use tomli for Python < 3.11

with open("pyproject.toml", "rb") as f:
    project_data = tomllib.load(f)


class ASPHandler(BaseHandler):
    """MKDocStrings handler for ASP files."""

    def __init__(
        self,
        theme: str = "material",
        **_kwargs: Any,
    ) -> None:
        """
        Initialize the handler.

        Args:
            theme: The theme to use for the handler.
            config_file_path: The path to the configuration file.
            paths: A list of paths to search for ASP files.
            locale: The locale to use for the handler.
            load_external_modules: Whether to load external modules.
            **kwargs: Keyword arguments.
        """
        super().__init__("asp", theme)
        self.env.filters["convert_markdown_simple"] = self.do_convert_markdown_simple

    def do_convert_markdown_simple(
        self,
        text: str,
        heading_level: int,
    ) -> Markup:
        """Render Markdown text without adding headers to the TOC

        Arguments:
            text: The text to convert.
            heading_level: The base heading level to start all Markdown headings from.

        Returns:
            An HTML string.
        """
        old_headings = [e for e in self._headings]
        treeprocessors = self._md.treeprocessors
        treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = heading_level  # type: ignore[attr-defined]

        try:
            md = Markup(self._md.convert(text))
        finally:
            treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = 0  # type: ignore[attr-defined]
            self._md.reset()

        self._headings = old_headings
        return md

    def parse_files(
        self, asp_parser: ASPParser, document_parser: DocumentParser, paths: list[Path]
    ) -> dict[Path, Document]:
        """
        Parse the files at the given paths and return a dictionary of documents.

        This also handles the inclusion of other files.

        Args:
            asp_parser: The ASP parser.
            document_parser: The document parser.
            paths: The paths to parse.

        Returns:
            A dictionary of documents.
        """
        parse_queue = deque(paths)
        documents: dict[Path, Document] = {}

        while parse_queue:
            path = parse_queue.popleft()
            if path.suffix != ".lp" or not path.is_file():
                continue

            with open(path, "r", encoding="utf-8") as f:
                content = f.read()

            tree = asp_parser.parse(content)
            document = document_parser.parse(Document(path, content), tree)
            documents[path] = document

            parse_queue.extend(include.path for include in document.includes if include.path not in documents)

        return documents

    def collect(self, identifier: str, config: dict) -> dict | None:
        """
        Collect data from ASP files.

        This function will be called for all markdown files annotated with '::: some/path/to/file.lp'.

        Args:
            identifier: The identifier used in the annotation.
            config: The configuration dictionary.

        Returns:
            The collected data as a dictionary.
        """

        # if identifier != "examples/my_test/base.lp":
        #     return None

        start_path = Path(identifier)
        asp_parser = ASPParser()
        document_parser = DocumentParser()

        documents: dict[Path, Document] = self.parse_files(asp_parser, document_parser, [start_path])

        data = {
            "project_name": project_data["project"]["name"],
            "project_url_tree": project_data["project"]["urls"]["Homepage"].replace(".git/", "/") + "tree/master/",
            "encodings": EncodingInfo.from_documents(documents.values()).encodings,
            "dependency_graph": DependencyGraph.from_document(documents.values()),
        }

        def get_key(x):
            key = x[1].show_status
            if x[1].show_status == 0:
                key = 3.5
            if not x[1].is_input:
                key += 0.1
            return key

        predicates = dict(sorted(PredicateInfo.from_documents(documents.values()).predicates.items(), key=get_key))
        data["predicate_info"] = predicates
        return data

    def render(self, data: dict, config: dict):
        """
        Render the collected data to html.

        This function will be called for all `data` collected by the collect function.

        Args:
            data: The data collected by the collect function.
            config: The configuration dictionary.

        Returns:
            The rendered data as a string
        """

        if data is None:
            return None

        if "start_level" not in config:
            config["start_level"] = 1

        # Get and render the documentation template
        template = self.env.get_template("documentation.html.jinja")
        # print("Rendering template with data:", data)
        return template.render(**data, config=config)

    def get_templates(self) -> Path:
        return Path(__file__).parent / "templates"



""" This module contains the DocumentParser class, which is responsible for parsing ASP documents.

    It extracts relevant information from the Tree-sitter parse tree and populates the Document object with
    statements, predicates, comments, and other elements.
"""

from tree_sitter import Node, Tree

from mkdocstrings_handlers.asp.document import Document
from mkdocstrings_handlers.asp.semantics.block_comment import BlockComment
from mkdocstrings_handlers.asp.semantics.directives.include import Include
from mkdocstrings_handlers.asp.semantics.directives.show import Show
from mkdocstrings_handlers.asp.semantics.line_comment import LineComment
from mkdocstrings_handlers.asp.semantics.predicate import Predicate
from mkdocstrings_handlers.asp.semantics.predicate_documentation import PredicateDocumentation
from mkdocstrings_handlers.asp.semantics.statement import Statement
from mkdocstrings_handlers.asp.tree_sitter.node_kind import NodeKind
from mkdocstrings_handlers.asp.tree_sitter.traverse import traverse


class DocumentParser:
    """
    A parser for ASP documents.
    """

    def __init__(self):
        """
        Initialize the parser.
        """
        self._reset()

    def _reset(self) -> None:
        """
        Reset the parser state.
        """
        self.head = True
        self.inside_tuple = False
        self.current_statement: Statement | None = None
        self.current_predicates: dict[str, Predicate] = {}
        self.error = False

    def parse(self, document: Document, tree: Tree) -> Document:
        """
        Parse the given tree.

        Args:
            tree: The tree to parse.
        """

        def _on_enter(node: Node) -> None:
            """
            Handle entering a node.

            Args:
                node: The node.
            """

            match NodeKind.from_grammar_name(node.grammar_name):
                # State management
                case NodeKind.HEAD:
                    self.head = True
                case NodeKind.BODY:
                    self.head = False
                case NodeKind.LITERAL_TUPLE:
                    self.inside_tuple = True

                # Data collection
                case NodeKind.STATEMENT:
                    self.current_statement = Statement.from_node(node)

                case NodeKind.SYMBOLIC_ATOM:
                    predicate = Predicate.from_node(node.parent)
                    signature = predicate.signature

                    if signature not in self.current_predicates:
                        self.current_predicates[signature] = predicate
                    else:
                        predicate = self.current_predicates[signature]

                    if self.head and not self.inside_tuple:
                        self.current_statement.add_provided(predicate)
                    else:
                        self.current_statement.add_needed(predicate)
                case NodeKind.LINE_COMMENT:
                    line_comment = LineComment.from_node(node)
                    document.line_comments.append(line_comment)
                    document.ordered_objects.append(line_comment)
                case NodeKind.BLOCK_COMMENT:
                    block_comment = BlockComment.from_node(node)
                    document.block_comments.append(block_comment)

                    # Predicate documentation
                    predicate_documentation = PredicateDocumentation.from_block_comment(block_comment)
                    if predicate_documentation is None:
                        document.ordered_objects.append(block_comment)
                        return

                    predicate = Predicate.from_node(predicate_documentation.node)
                    signature = predicate.signature

                    if signature not in document.predicates:
                        document.predicates[signature] = predicate
                    else:
                        predicate = document.predicates[signature]

                    predicate.documentation = predicate_documentation
                    predicate.documentation.node = None
                case NodeKind.SHOW_SIGNATURE | NodeKind.SHOW | NodeKind.SHOW_TERM:
                    show = Show.from_node(node)
                    if show:
                        predicate = show.predicate
                        if predicate.signature in self.current_predicates:
                            predicate = self.current_predicates[predicate.signature]
                        else:
                            self.current_predicates[predicate.signature] = predicate

                        predicate.update_show_status(show.predicate.show_status)

                        # If the show directive is not empty
                        # default show behaviour depends on the show directive
                        document.disable_default_show |= show.disable_default

                        # We regard the found predicate as a provided predicate
                        # in order to track its dependencies
                        self.current_statement.add_provided(predicate)
                    else:
                        # If show returns None
                        # this means that the show directive was empty
                        # and we should disable the default show behaviour
                        document.disable_default_show = True

                case NodeKind.INCLUDE:
                    include = Include.from_node(document.path, node)
                    document.includes.append(include)
                case NodeKind.ERROR:
                    self.error = True
                case _:
                    pass

        def _on_exit(node: Node) -> None:
            """
            Handle exiting a node.

            Args:
                node: The node.
            """
            match NodeKind.from_grammar_name(node.grammar_name):
                case NodeKind.STATEMENT:
                    if self.error:
                        self._reset()
                        return

                    # The statement is done
                    document.statements.append(self.current_statement)
                    document.ordered_objects.append(self.current_statement)

                    for signature, predicate in self.current_predicates.items():
                        if signature not in document.predicates:
                            document.predicates[signature] = predicate
                        else:
                            document.predicates[signature].update_show_status(predicate.show_status)

                    self._reset()

                case NodeKind.LITERAL_TUPLE:
                    self.inside_tuple = False
                case _:
                    pass

        self._reset()
        traverse(tree, _on_enter, _on_exit)

        return document



from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Dict

from mkdocstrings_handlers.asp.document import Document
from mkdocstrings_handlers.asp.semantics.predicate import Predicate, ShowStatus


@dataclass
class PredicateInfo:
    predicates: Dict[str, Predicate] = field(default_factory=OrderedDict)
    """Dictionary of predicates with their signatures as keys."""

    @staticmethod
    def from_documents(documents: dict[Document]) -> PredicateInfo:
        """
        Create a predicate list from a list of documents.

        Args:
            documents: The documents to create the predicate list from.

        Returns:
            The predicate list.
        """
        predicates: dict[str, Predicate] = {}
        disable_default_show = False

        # If any document has disable_default_show set to True
        # set it to True for all predicates
        for document in documents:
            disable_default_show |= document.disable_default_show

        # Collect all predicates with updated show status
        for document in documents:
            for predicate in document.predicates.values():
                signatuture = predicate.signature
                if signatuture not in predicates:
                    predicates[signatuture] = predicate

                if disable_default_show:
                    predicates[signatuture].update_show_status(ShowStatus.HIDDEN)

                predicates[signatuture].update_show_status(predicate.show_status)
                if predicates[signatuture].documentation is None:
                    predicates[signatuture].documentation = predicate.documentation

        # Create an ordered dictionary sorted by signature
        sorted_predicates = OrderedDict(sorted(predicates.items(), key=lambda x: x[0]))

        return PredicateInfo(sorted_predicates)



"""
Test cases for main application functionality.
"""

from io import StringIO
from unittest import TestCase

from mkdoclingo.utils import logging
from mkdoclingo.utils.logging import configure_logging, get_logger
from mkdoclingo.utils.parser import get_parser


class TestMain(TestCase):
    """
    Test cases for main application functionality.
    """

    def test_logger(self) -> None:
        """
        Test the logger.
        """
        sio = StringIO()
        configure_logging(sio, logging.INFO, True)
        log = get_logger("main")
        log.info("test123")
        self.assertRegex(sio.getvalue(), "test123")

    def test_parser(self) -> None:
        """
        Test the parser.
        """
        parser = get_parser()
        ret = parser.parse_args(["--log", "info"])
        self.assertEqual(ret.log, logging.INFO)



""" This module contains the representation of an include directive in ASP."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from tree_sitter import Node


@dataclass
class Include:
    """An include directive in an ASP document."""

    path: Path
    """ The path of the included file."""

    @staticmethod
    def from_node(path: Path, node: Node) -> Include:
        """
        Create an Include from the given node.

        Args:
            node: The node.

        Returns:
            The created Include.
        """
        # If the node is an include,
        # then the first child is the include directive
        # and the second child is the file path.

        # The second child of the file path is the file path
        # as a string fragment without the quotes.
        file_path_node = node.children[1]
        file_path = Path(file_path_node.children[1].text.decode("utf-8"))

        # return Include(file_path)
        return Include(path.parent.joinpath(file_path))



"""
The command line parser for the project.
"""

from argparse import ArgumentParser
from importlib import metadata
from textwrap import dedent
from typing import Any, Optional, cast

from . import logging

__all__ = ["get_parser"]

VERSION = metadata.version("mkdoclingo")


def get_parser() -> ArgumentParser:
    """
    Return the parser for command line options.
    """
    parser = ArgumentParser(
        prog="mkdoclingo",
        description=dedent(
            """\
            mkdoclingo
            filldescription
            """
        ),
    )
    levels = [
        ("error", logging.ERROR),
        ("warning", logging.WARNING),
        ("info", logging.INFO),
        ("debug", logging.DEBUG),
    ]

    def get(levels: list[tuple[str, int]], name: str) -> Optional[int]:
        for key, val in levels:
            if key == name:
                return val
        return None  # nocoverage

    parser.add_argument(
        "--log",
        default="warning",
        choices=[val for _, val in levels],
        metavar=f"{{{','.join(key for key, _ in levels)}}}",
        help="set log level [%(default)s]",
        type=cast(Any, lambda name: get(levels, name)),
    )

    parser.add_argument("--version", "-v", action="version", version=f"%(prog)s {VERSION}")
    return parser



"""This module contains the Document class, which represents a single document in the context of ASP."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from mkdocstrings_handlers.asp.semantics.block_comment import BlockComment
from mkdocstrings_handlers.asp.semantics.directives.include import Include
from mkdocstrings_handlers.asp.semantics.line_comment import LineComment
from mkdocstrings_handlers.asp.semantics.predicate import Predicate
from mkdocstrings_handlers.asp.semantics.statement import Statement


@dataclass
class Document:
    """
    A document representing the content of a particular ASP file.
    """

    path: Path
    content: str
    statements: list[Statement] = field(default_factory=list)
    line_comments: list[LineComment] = field(default_factory=list)
    block_comments: list[BlockComment] = field(default_factory=list)
    ordered_objects: list[Statement | LineComment | BlockComment] = field(default_factory=list)
    predicates: dict[str, Predicate] = field(default_factory=dict)
    includes: list[Include] = field(default_factory=list)
    disable_default_show: bool = False



"""This module contains the 'PredicateDocumentation', which represents the documentation for a predicate."""

from __future__ import annotations

import textwrap
from dataclasses import dataclass, field

from tree_sitter import Node

from mkdocstrings_handlers.asp.semantics.block_comment import BlockComment
from mkdocstrings_handlers.asp.tree_sitter.node_kind import NodeKind
from mkdocstrings_handlers.asp.tree_sitter.parser import ASPParser
from mkdocstrings_handlers.asp.tree_sitter.traverse import traverse


@dataclass
class PredicateDocumentation:
    """
    Documentation for a predicate.


    Example:
        %*#some_predicate(A,B,C).
        description
        #parameters
        - A : this is  A
        - B : this is  B
        - C : this is  C
        *%
    """

    signature: str
    """ The signature of the predicate. """
    description: str
    """ The description of the predicate. """
    parameter_descriptions: dict[str, str] = field(default_factory=dict)
    """ The descriptions of the parameters of the predicate. """
    node: Node | None = None
    """ The node representing the predicate. """

    @staticmethod
    def from_block_comment(comment: BlockComment) -> PredicateDocumentation | None:
        """
        Create a predicate documentation from a comment.

        Args:
            comment: The block comment.

        Returns:
            The predicate documentation or None if the comment is not a predicate documentation.
        """
        if not comment.lines[0].startswith("#"):
            return None

        # Get the signature
        signature = comment.lines[0].removeprefix("#").strip()

        # Parse the signature to get the literal
        predicate_node = None

        def identifier_from_node(node: Node):
            nonlocal predicate_node
            if NodeKind.from_grammar_name(node.grammar_name) == NodeKind.SYMBOLIC_ATOM:
                predicate_node = node.parent

        parser = ASPParser()
        tree = parser.parse(signature)
        traverse(tree, identifier_from_node, lambda _: None)

        # Get the description
        description_lines = []

        for line in comment.lines[1:]:
            if line.startswith("#parameters"):
                description = line.removeprefix("#").strip()
                break
            description_lines.append(line)

        description = "\n".join(description_lines)
        parameters = []
        parameter_descriptions = {}
        for line in comment.lines[len(description_lines) :]:
            if line.startswith("-"):
                parts = line.removeprefix("-").split(":")
                if len(parts) == 2:
                    parameter, parameter_description = parts
                    parameters.append(parameter.strip())
                    parameter_descriptions[parameter.strip()] = parameter_description.strip()
        description = textwrap.dedent(description).strip()
        return PredicateDocumentation(
            signature=signature.removesuffix("."),
            description=description,
            parameter_descriptions=parameter_descriptions,
            node=predicate_node,
        )



""" This module contains the representation of a block or multi-line comment in an ASP document."""

from __future__ import annotations

from dataclasses import dataclass

from tree_sitter import Node


@dataclass
class BlockComment:
    """A block comment in an ASP document."""

    row: int
    """ The row of the block comment. """
    lines: list[str]
    """ The lines of text of the block comment. """

    @staticmethod
    def from_node(node: Node) -> BlockComment:
        """
        Create a block comment from a node.
        """
        clean_text = node.text.decode("utf-8").removeprefix("%*").removesuffix("*%").strip()
        lines = clean_text.split("\n")

        return BlockComment(row=node.start_point.row, lines=lines)
